# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Author: Ole-Johan Skrede                                                    #
# Creation date: 14.10.2016                                                   #
# Revised: 02.11.2016                                                         #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 9, task 2.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation

import os
import sys

import numpy as np
import cv2
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import seaborn as sns # pylint: disable=import-error,unused-import

# Import module from week_07
path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../week_07'))
if not path in sys.path:
  sys.path.insert(1, path)
del path
import multivariate_gauss_classifier#pylint:disable=import-error,wrong-import-position

def plot_image(image, fig_num, colormap='gray'):
  """Plot image"""
  #plt.figure(fig_num)
  _, ax = plt.subplots()
  im = ax.imshow(image, cmap=colormap, interpolation='none')
  #ax.set_xticks(np.arange(0, image.shape[1]))
  ax.set_xticks(np.arange(0.5, image.shape[1] + 0.5), minor=True)
  ax.get_xaxis().set_ticks([]) # Hide tick labels
  ax.set_xlabel(r'$x_1$', fontsize=16) # Latex rendering
  #ax.set_yticks(np.arange(0, image.shape[0]))
  ax.set_yticks(np.arange(0.5, image.shape[0] + 0.5), minor=True)
  ax.get_yaxis().set_ticks([]) # Hide tick labels
  ax.set_ylabel(r'$x_2$', fontsize=16)
  ax.grid(b=False, which='major')
  ax.grid(b=True, which='minor', linestyle='-', color='gray')
  plt.xticks(np.arange(0, 21), np.arange(-10, 11))
  plt.yticks(np.arange(0, 21), np.arange(10, -11, -1))
  plt.colorbar(im)

  plt.tight_layout()

  fig_num += 1
  return fig_num

def plot_contours(fig_num, matrix, xmin, xmax, ymin, ymax, levels, vmin, vmax, emph_zero=False):
  """Contour plot"""
  plt.figure(fig_num)
  matplotlib.rcParams['contour.negative_linestyle'] = 'solid'
  im = plt.imshow(matrix,
                  interpolation='bilinear', origin='lower',
                  cmap='magma', extent=[xmin, xmax, ymin, ymax],
                  vmin=vmin, vmax=vmax)
  CS = plt.contour(matrix, levels, origin='lower', linewidths=2,
                   colors='white', extent=[xmin, xmax, ymin, ymax])

  if emph_zero:
    # Thicken and dash zero contour (which is the decision boundary)
    zc = CS.collections[10] # You must know the index (in levels) of the zero contour
    plt.setp(zc, linewidth=4, linestyle='--')

  plt.clabel(CS, levels[0::2], inline=1, fmt='%1.0f', fontsize=14)
  plt.colorbar(im)
  plt.xlabel(r'$x_1$', fontsize=16)
  plt.ylabel(r'$x_2$', fontsize=16)
  plt.grid('off')
  plt.tight_layout()

  fig_num += 1
  return fig_num

def plot_class_image(image, fig_num, classes=None, name=None, colormap='gray',
                     write_file=None, pixelated=False):
  """Plot class image"""
  if classes is None:
    classes = np.unique(image)

  if not isinstance(classes, list):
    classes = list(classes)

  cmap = plt.cm.get_cmap(colormap)

  patches = []
  for c in classes:
    color = cmap(c/np.max(classes))
    classname = 'C '+str(c)
    patches.append(mpatches.Patch(color=color, label=classname))

  if pixelated: # Display pixel boundaries
    _, ax = plt.subplots()
    ax.imshow(image, cmap=colormap, interpolation='none', vmin=0,
              vmax=classes[-1])
    #ax.set_xticks(np.arange(0, image.shape[1]))
    ax.set_xticks(np.arange(0.5, image.shape[1] + 0.5), minor=True)
    ax.get_xaxis().set_ticks([]) # Hide tick labels
    ax.set_xlabel(r'$x_1$', fontsize=16)
    #ax.set_yticks(np.arange(0, image.shape[0]))
    ax.set_yticks(np.arange(0.5, image.shape[0] + 0.5), minor=True)
    ax.get_yaxis().set_ticks([]) # Hide tick labels
    ax.set_ylabel(r'$x_2$', fontsize=16)
    ax.grid(b=False, which='major')
    ax.grid(b=True, which='minor', linestyle='-', color='gray')
    plt.xticks(np.arange(0, 21), np.arange(-10, 11))
    plt.yticks(np.arange(0, 21), np.arange(10, -11, -1))
    plt.legend(bbox_to_anchor=(0, 1.02, 1., .102), loc=3, ncol=len(classes),
               mode="expand", borderaxespad=0, handles=patches)

  else:
    plt.figure(fig_num)
    plt.imshow(image, cmap=colormap, interpolation='none', vmin=0, vmax=classes[-1])
    if name is not None:
      plt.title(name)
    plt.xticks([]), plt.yticks([])
    plt.legend(bbox_to_anchor=(0, 1.02, 1., .102), loc=3, ncol=len(classes),
               mode="expand", borderaxespad=0, handles=patches)

  if write_file:
    cv2.imwrite(write_file, image)
  fig_num += 1
  return fig_num

def main():
  """main"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 8')
  print('Task 1')
  print('-'*80)

  subtasks = ['a', 'c', 'd', 'e']
  fig_num = 0

  classes = np.array([1, 2])
  prior = np.array([0.5, 0.5])
  class_means = np.array([[3, 6], [3, -2]])
  class_covs = np.array([[[0.5, 0], [0, 2]], [[2, 0], [0, 2]]])
  class_inv_covs = np.array([[[2, 0], [0, 0.5]], [[0.5, 0], [0, 0.5]]])


  # Task 1 a)
  if 'a' in subtasks:
    # Eigenvalues and eigenvectors of the covariance matrix
    for c in range(2):
      eig_vals, eig_vecs = np.linalg.eig(class_covs[c, :, :])
      print("Eigenvalues for class {}: {}".format(classes[c], eig_vals))
      print("Eigenvectors for class {}:".format(classes[c]))
      print(eig_vecs)

    delta = 0.1
    xmin = -8
    xmax = 10
    ymin = -8
    ymax = 10
    vmin = -130
    vmax = 40
    x_vals = np.arange(xmin, xmax, delta)
    y_vals = np.arange(ymin, ymax, delta)
    xx, _ = np.meshgrid(x_vals, y_vals)
    discr = np.zeros((2, xx.shape[0], xx.shape[1]))
    for i, x in enumerate(x_vals):
      for j, y in enumerate(y_vals):
        for c in range(2):
          mean_diff = np.array([x, y]) - class_means[c, :]
          # Note j, i because of pyplot imshow convention
          discr[c, j, i] = (np.log(prior[c]) -
                            1/2*np.log(np.linalg.det(class_covs[c, :, :])) -
                            1/2*np.dot(mean_diff.T, np.dot(class_inv_covs[c, :, :], mean_diff)))


    # Separate
    levels = [np.arange(-100, 0, 10), np.arange(-60, 0, 10)]
    for c in range(2):
      fig_num = plot_contours(fig_num, discr[c, :, :], xmin, xmax, ymin, ymax,
                              levels[c], vmin, vmax)
      plt.plot(class_means[c, 0], class_means[c, 1], 'ko')
      plt.xlim([xmin, xmax])
      plt.ylim([ymin, ymax])

    # Both (all points < 0 belongs to class 2)
    fig_num = plot_contours(fig_num, discr[0, :, :] - discr[1, :, :], xmin, xmax,
                            ymin, ymax, np.arange(-100, 40, 10), vmin, vmax, emph_zero=True)
    plt.plot(class_means[0, 0], class_means[0, 1], 'ko')
    plt.plot(class_means[1, 0], class_means[1, 1], 'ko')
    plt.xlim([xmin, xmax])
    plt.ylim([ymin, ymax])


  # Task 1 c)
  if 'c' and 'a' in subtasks:
    # Plot decision boundary on top of existing contour-plot. This line should
    # lie on top of the zero-contour.
    x1 = np.linspace(-3.60233769, 9.60233769, 1000)
    x2 = 59/16 - np.log(2)/4 - 9/8*x1 + 3/16*x1*x1
    plt.plot(x1, x2, c="k")
    plt.xlim([xmin, xmax])
    plt.ylim([ymin, ymax])

  # Task 1 d)
  if 'd' in subtasks:
    feature_images = np.zeros((2, 21, 21))
    for i in range(21):
      for j in range(21):
        x = -10 + i
        y = -10 + j

        feature_images[0, i, j] = y
        feature_images[1, i, j] = x


    fig_num = plot_image(np.flipud(feature_images[0, :, :]), fig_num,
                         colormap='inferno')
    fig_num = plot_image(np.flipud(feature_images[1, :, :]), fig_num,
                         colormap='inferno')

  # Task 1 e)
  if 'd' and 'e' in subtasks:
    # Classify an image based on the two feature images from 'd', both with the
    # analytically computed decision boundary, the multivariate gaussian
    # classifier from last week, and the k-nearest-neighbour classifier which
    # we implemented in Task 2 this week.

    classes = np.array([1, 2])

    # Analytical
    classified_anal = np.zeros((21, 21))
    for i in range(21):
      for j in range(21):
        x1 = feature_images[0, i, j]
        x2 = feature_images[1, i, j]
        if x2 > 59/16 - np.log(2)/4 - 9/8*x1 + 3/16*x1*x1:
          classified_anal[i, j] = classes[0]
        else:
          classified_anal[i, j] = classes[1]

    fig_num = plot_class_image(np.flipud(classified_anal), fig_num, classes=classes,
                               colormap='rainbow', pixelated=True)

    # Multivariate gauss
    mvgc = multivariate_gauss_classifier.Classification()
    mvgc.classify([feature_images[0, :, :], feature_images[1, :, :]],
                  class_means, class_covs, classes)
    classified_gauss = mvgc.prop_classified_image
    fig_num = plot_class_image(np.flipud(classified_gauss), fig_num, classes=classes,
                               colormap='rainbow', pixelated=True)




  plt.show()


if __name__ == "__main__":
  main()
